@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -40,10 +40,12 @@
 
 #include "device.h"
 #include "server.h"
+#include "manager.h"
 
 static int idle_timeout = 0;
 
 static DBusConnection *connection = NULL;
+static GSList *adapters = NULL;
 
 static void input_remove(struct btd_device *device, const char *uuid)
 {
@@ -69,7 +71,7 @@ static int hid_device_probe(struct btd_device *device, GSList *uuids)
 	adapter_get_address(adapter, &src);
 	device_get_address(device, &dst);
 
-	return input_device_register(connection, path, &src, &dst,
+	return input_device_register(connection, device, path, &src, &dst,
 				HID_UUID, rec->handle, idle_timeout * 60);
 }
 
@@ -82,13 +84,19 @@ static int headset_probe(struct btd_device *device, GSList *uuids)
 {
 	struct btd_adapter *adapter = device_get_adapter(device);
 	const gchar *path = device_get_path(device);
-	const sdp_record_t *record = btd_device_get_record(device, uuids->data);
+	const sdp_record_t *record;
 	sdp_list_t *protos;
 	uint8_t ch;
 	bdaddr_t src, dst;
 
 	DBG("path %s", path);
 
+	if (!g_slist_find_custom(uuids, HSP_HS_UUID,
+					(GCompareFunc) strcasecmp))
+		return -EINVAL;
+
+	record = btd_device_get_record(device, uuids->data);
+
 	if (!record || sdp_get_access_protos(record, &protos) < 0) {
 		error("Invalid record");
 		return -EINVAL;
@@ -106,7 +114,7 @@ static int headset_probe(struct btd_device *device, GSList *uuids)
 	adapter_get_address(adapter, &src);
 	device_get_address(device, &dst);
 
-	return fake_input_register(connection, path, &src, &dst,
+	return fake_input_register(connection, device, path, &src, &dst,
 				HSP_HS_UUID, ch);
 }
 
@@ -118,10 +126,17 @@ static void headset_remove(struct btd_device *device)
 static int hid_server_probe(struct btd_adapter *adapter)
 {
 	bdaddr_t src;
+	int ret;
 
 	adapter_get_address(adapter, &src);
 
-	return server_start(&src);
+	ret = server_start(&src);
+	if (ret < 0)
+		return ret;
+
+	adapters = g_slist_append(adapters, btd_adapter_ref(adapter));
+
+	return 0;
 }
 
 static void hid_server_remove(struct btd_adapter *adapter)
@@ -131,6 +146,9 @@ static void hid_server_remove(struct btd_adapter *adapter)
 	adapter_get_address(adapter, &src);
 
 	server_stop(&src);
+
+	adapters = g_slist_remove(adapters, adapter);
+	btd_adapter_unref(adapter);
 }
 
 static struct btd_device_driver input_hid_driver = {
